//==============================================================================
//	
//	Copyright (c) 2011-
//	Authors:
//	* Christian Krause <christian.krause@hpi.uni-potsdam.de> (Hasso Plattner Institute)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package ipta;

import prism.PrismException;
import pta.Constraint;
import pta.Edge;
import pta.PTA;
import pta.Transition;

public class IntervalTransition extends Transition {
	
	/**
	 * Default constructor.
	 * @param parent IPTA.
	 * @param src Source location.
	 * @param action Action.
	 */
	public IntervalTransition(IPTA parent, int src, String action) 
	{
		super(parent, src, action);
	}
	
	/**
	 * Copy this interval transition.
	 */
	@Override
	public IntervalTransition copy()
	{
		IntervalTransition copy = new IntervalTransition((IPTA) parent, src, action);
		for (Constraint constraint : guard) {
			copy.addGuardConstraint(new Constraint(constraint));
		}
		for (Edge edge : edges) {
			copy.addEdge(edge.copy());
		}
		return copy;
	}
	
	@Override
	public void setParent(PTA parent)
	{
		if (!(parent instanceof IPTA)) {
			throw new RuntimeException("Parent must be an IPTA");
		} else {
			super.setParent(parent);
		}
	}

	@Override
	public IPTA getParent()
	{
		return (IPTA) parent;
	}
	
	public IntervalEdge addEdge(double lower, double upper, int dest)
	{
		IntervalEdge edge = new IntervalEdge(this, lower, upper, dest);
		super.addEdge(edge);
		return edge;
	}

	@Override
	public Edge addEdge(double prob, int dest)
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public void addEdge(Edge e)
	{
		if (e instanceof IntervalEdge) {
			super.addEdge(e);
		} else {
			throw new UnsupportedOperationException();			
		}
	}
	
	@Override
	public void check() throws PrismException
	{
		// sum of the lower bounds must be <= 1
		// sum of the upper bounds must be >= 1
		double sumLower = 0.0;
		double sumUpper = 0.0;
		double err = 1e-10;
		for (Edge e : edges) {
			e.check();
			sumLower += ((IntervalEdge) e).getLowerBound();
			sumUpper += ((IntervalEdge) e).getUpperBound();
		}
		if (sumLower > 1.0+err) {
			throw new PrismException("Sum of lower bounds must be <= 1 for IPTA transition \""+this+"\"");
		}
		if (sumUpper < 1.0-err) {
			throw new PrismException("Sum of upper bounds must be >= 1 for IPTA transition \""+this+"\"");
		}
	}
	
}
