//==============================================================================
//	
//	Copyright (c) 2011-
//	Authors:
//	* Christian Krause <christian.krause@hpi.uni-potsdam.de> (Hasso Plattner Institute)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package ipta;

import parser.Values;
import parser.ast.Expression;
import parser.ast.ExpressionBinaryOp;
import parser.ast.ExpressionUnaryOp;
import parser.ast.ModulesFile;
import prism.ModelType;
import prism.Prism;
import prism.PrismLangException;
import pta.Edge;
import pta.Modules2PTA;
import pta.PTA;
import pta.Transition;

public class Modules2IPTA extends Modules2PTA {
	
	/**
	 * Default constructor.
	 * @param prism Prism core.
	 * @param modulesFile Modules file.
	 */
	public Modules2IPTA(Prism prism, ModulesFile modulesFile) {
		super(prism, modulesFile);
	}
	
	/*
	 * (non-Javadoc)
	 * @see pta.Modules2PTA#translate()
	 */
	@Override
	public IPTA translate() throws PrismLangException
	{
		return (IPTA) super.translate();
	}
	
	/*
	 * (non-Javadoc)
	 * @see pta.Modules2PTA#checkModulesFile()
	 */
	@Override
	protected void checkModulesFile() throws PrismLangException
	{
		if (modulesFile.getModelType() != ModelType.IPTA)
			throw new PrismLangException("Model is not an IPTA");
		if (modulesFile.getNumGlobals() > 0)
			throw new PrismLangException("IPTA models cannot have global variables");
		if (modulesFile.getInitialStates() != null)
			throw new PrismLangException("IPTA models cannot use init...endinit");
	}

	/*
	 * (non-Javadoc)
	 * @see pta.Modules2PTA#createPTA()
	 */
	@Override
	protected PTA createPTA() {
		return new IPTA();
	}

	/*
	 * (non-Javadoc)
	 * @see pta.Modules2PTA#parallel(pta.PTA, pta.PTA)
	 */
	@Override
	protected PTA parallel(PTA pta1, PTA pta2) {
		return new IPTAParallel().compose(pta1, pta2);
	}

	/*
	 * (non-Javadoc)
	 * @see pta.Modules2PTA#createEdge(pta.Transition, parser.ast.Expression, parser.Values, int)
	 */
	@Override
	protected Edge createEdge(Transition transition, Expression expression, Values values, int dest) throws PrismLangException {
		
		// Get the lower and upper bounds:
		double lowerBound = 1.0;
		double upperBound = 1.0;
		
		if (expression!=null) {
			
			// Ignore the outer parentheses.
			while (expression instanceof ExpressionUnaryOp &&
					((ExpressionUnaryOp) expression).getOperator()==ExpressionUnaryOp.PARENTH) {
				expression = ((ExpressionUnaryOp) expression).getOperand();
			}
			
			// Interval or point?
			if (expression instanceof ExpressionBinaryOp &&
				((ExpressionBinaryOp) expression).getOperator()==ExpressionBinaryOp.INTERVAL) {
					ExpressionBinaryOp binary = (ExpressionBinaryOp) expression; 
					lowerBound = binary.getOperand1().evaluateDouble(constantValues, values);
					upperBound = binary.getOperand2().evaluateDouble(constantValues, values);
			} else {
				lowerBound = expression.evaluateDouble(constantValues, values);
				upperBound = lowerBound;
			}
		}
		
		// Create the new edge:
		return ((IntervalTransition) transition).addEdge(lowerBound, upperBound, dest);
		
	}

}
